// I2C EEPROM reader / programmer
// by SP

// Arduino creates interface between I2CMprog running on PC with Windows
// and I2C serial EEPROM memory connected to its +5V, GND, SDA, SCL

// Created 30 July 2017

// This code is in the public domain.


#include <Wire.h>

char cmd_buffer[135];  // buffer for data read/write
char mem_type;
char mem_size;
char mem_mask;

void setup()
{
  Wire.begin();        // join i2c bus (address optional for master)
  Serial.begin(38400);  // start serial for output
  Serial.setTimeout(500);
}

 
// Return: 0 - success; >0 - error
unsigned char ping_i2c(unsigned char slave_addr) 
{
  Wire.beginTransmission(slave_addr); // transmit to device
  return(Wire.endTransmission());    // stop transmitting
  delay(5);
}

// read 1 byte from short address
byte read_i2c( int busAdr, byte adrRead ) 
{
  byte rdata = 0xFF;
  // First dummy write (for cheating 2 bytes address memory)
  Wire.beginTransmission(busAdr);
  Wire.write(adrRead); // LSB
  Wire.endTransmission();
  // And standard reading now
  Wire.beginTransmission(busAdr);
  Wire.write(adrRead); // LSB
  Wire.endTransmission();
  delay(5);
  Wire.requestFrom(busAdr,1);
  if (Wire.available()) rdata = Wire.read();
  return rdata;
}

// read 1 byte from long address
byte readL_i2c( int busAdr, unsigned int adrRead ) 
{
  byte rdata = 0xFF;
  Wire.beginTransmission(busAdr);
  Wire.write((byte)(adrRead >> 8)); // MSB
  Wire.write((byte)(adrRead & 0xFF)); // LSB
  Wire.endTransmission();
  Wire.requestFrom(busAdr,1);
  if (Wire.available()) rdata = Wire.read();
  return rdata;
}

// write byte with 2 address bytes
void writeL_i2c(byte busAdr, unsigned int adrWrt, byte data ) 
{
  Wire.beginTransmission(busAdr);
  Wire.write((int)(adrWrt >> 8)); // MSB
  Wire.write((int)(adrWrt & 0xFF)); // LSB
  Wire.write(data);
  Wire.endTransmission();
}

// write byte with 1 address byte
void write_i2c(byte busAdr, byte adrWrt, byte data ) 
{
  Wire.beginTransmission(busAdr);
  Wire.write(adrWrt); // LSB
  Wire.write(data);
  Wire.endTransmission();
}

// test address - if is valid return true
boolean test_adr(unsigned char tstAdr) 
{
  byte mem_H,mem_L,mem_N;
  mem_H = read_i2c(0x50, tstAdr);
  mem_L = read_i2c(0x50, (tstAdr>>1));
  if(mem_H != mem_L) return true;
  write_i2c(0x50, tstAdr, (mem_H ^ 0xA5));
  delay(10);
  mem_N = read_i2c(0x50, (tstAdr>>1));
  write_i2c(0x50, tstAdr, mem_H);
  delay(10);
  if(mem_N == mem_L) return true; else return false; 
}

// test long address - if is valid return true
boolean test_Ladr(unsigned int tstAdr) 
{
  byte mem_H,mem_L,mem_N;
  mem_H = readL_i2c(0x50, tstAdr);
  mem_L = readL_i2c(0x50, (tstAdr>>1));
  if(mem_H != mem_L) return true;
  writeL_i2c(0x50, tstAdr, (mem_H ^ 0xA5));
  delay(10);
  mem_N = readL_i2c(0x50, (tstAdr>>1));
  writeL_i2c(0x50, tstAdr, mem_H);
  delay(10);
  if(mem_N == mem_L) return true; else return false; 
}

void loop()
{
  byte auxbyte,auxbyte2,adres,comnd,i,j;
  unsigned int lAdres;

  // check if data has been sent from the computer:
  if (Serial.available()) {
    // read the command byte (which will be upper chars 'V','P','Z','W','R',T' or 'z'):
    comnd = Serial.read(); // read the command
    // "ping" command to i2c address
    if (comnd == 'P') {
      if(Serial.readBytes(cmd_buffer,1) == 1) {
        if(ping_i2c(cmd_buffer[0])) Serial.write('N'); else Serial.print('Y');
      }
    }

    // "zero" 1-byte memory counter in the i2c memory
    if (comnd == 'V') {
      Serial.write('1');
    }

    // "zero" 1-byte memory counter in the i2c memory
    if (comnd == 'z') {
      Wire.beginTransmission(0x50);
      Wire.write(0);
      Wire.endTransmission();
      Serial.print('z');
    }
    
    // "zero" 2-byte memory counter in the i2c memory
    if (comnd == 'Z') {
      Wire.beginTransmission(0x50);
      Wire.write(0);
      Wire.write(0);
      Wire.endTransmission();
      Serial.print('Z');
    }
        
    // "write" command to i2c address
    if (comnd == 'W') {
      if(Serial.readBytes(cmd_buffer,1) == 1) {
        i=cmd_buffer[0];
          if(i && (Serial.readBytes(cmd_buffer,i) == i)) {
            Wire.beginTransmission(cmd_buffer[0]);
            for(j=1; j<i; j++) Wire.write(cmd_buffer[j]);
            Wire.endTransmission();
            Serial.print('A');
          } else Serial.print('E');
      } else Serial.print('E');
    }
    
    // "read" command from i2c address
    if (comnd == 'R') {
      if(Serial.readBytes(cmd_buffer,1) == 1) {
        adres=cmd_buffer[0];
          if(Serial.readBytes(cmd_buffer,1) == 1) {
            i=cmd_buffer[0];
            j=0;
            Wire.requestFrom(adres, i);    // request i bytes from slave device adres
            while(Wire.available())    // slave may send less than requested
            {
              cmd_buffer[j++] = Wire.read(); // receive a byte as character
            }
            for(i=0; i<j; i++) Serial.print(cmd_buffer[i]);
          }
      }
    }

    // "test" command - try to identify memory on i2c bus
    if (comnd == 'T') {
      mem_type = 0;
      mem_size = 1;
      mem_mask = 0;
      // try to detect such exotic memory as A24C01 (without bus address)
      if(!ping_i2c(1) && !ping_i2c(0x78)) {
        // something on the bus (or short circuit :-), so test last memory address
        Wire.requestFrom(0x7F,1);
        if (Wire.available()) i = Wire.read();
        // write reverse byte
        Wire.beginTransmission(0x7F);
        Wire.write(~i);
        Wire.endTransmission();
        // and test writing
        Wire.requestFrom(0x7F,1);
        if (Wire.available()) j = Wire.read();
        if (j == (~i)) {
          // restore data
          Wire.beginTransmission(0x7F);
          Wire.write(i);
          Wire.endTransmission();
          mem_type = 1;
          mem_size = 7;
        }
      } else {
        if(!ping_i2c(0x50)) {
          delay(10);
          // There is something with 1010000 address on the bus
          // 1. test address mode (2 or 1 adr byte)
          // a) Read value from address 0 with 1 address byte only
          auxbyte = read_i2c(0x50,0);
          // b) Write value with 4 changed bytes (use only 1 address byte)
          write_i2c(0x50,0,auxbyte^0xA5);
          delay(10);
          // c) Check if writing with 1 address byte is working
          i = read_i2c(0x50,0);
          if(auxbyte == (i^0xA5)) mem_type = 2; else mem_type = 3;
          // d) and save original value
          if(mem_type == 2) write_i2c(0x50,0,auxbyte);
          delay(10);
          // Test memory size
          if(mem_type == 2) {
            // 1 address byte memory
            j = 0xFF; i=8;
            do {
              if(test_adr(j)) mem_size = i;
              else {i--; j=(j>>1);}
            } while ((i>3) && (mem_size<3));
            } else {
            // 2 address byte memory
            lAdres = 0xFFFF; i=8;
            do {
              if(test_Ladr(lAdres)) mem_size = i;
              else {i--; lAdres=(lAdres>>1);}
            } while ((i>3) && (mem_size<3));
          }
          // test block mask
          if(mem_size == 8) {
            delay(10);
            if(!ping_i2c(0x51)) mem_mask |= 0x01;
            delay(10);
            if(!ping_i2c(0x52)) mem_mask |= 0x02;
            delay(10);
            if(!ping_i2c(0x54)) mem_mask |= 0x04;
            delay(10);
          }
        }
      }
      Serial.print((byte)mem_type);         // print memory type
      Serial.print((byte)mem_size);         // print memory size
      Serial.print((byte)mem_mask);         // print memory block mask
    }
  }
  delay(10);
}
